%{
#include <stdarg.h>

static void echof(const char* format, ...);
static void echo_color(const char* color);
static void enter(int cond);
static void leave(void);

#define CONDSTACK_SIZE  128

static int cond_stack[CONDSTACK_SIZE];
static int stack_ptr;
%}

%x STRING
%x CHARA
%x FILENAME
%x COMMENT
%x LINE_COM
%s INCLUDE
%s PRAGMA

NONZERO		[1-9]
DIGIT		[[:digit:]]
XDIGIT		[[:xdigit:]]
FLOAT		({NONZERO}{DIGIT}*\.?|0\.|\.{DIGIT}){DIGIT}*|0
EXPONENT	[eE](-|\+?){DIGIT}+

TYPE		void|char|int|float|double|short|long|signed|unsigned|sizeof
STO_CLASS	extern|static|auto|register
TYPEDEF		typedef|enum|struct|union
CONTROL		if|else|for|while|do|switch|case|default|goto|return|break|continue
TYPE_QUAL	const|volatile
C_KEYWORDS	{TYPE}|{STO_CLASS}|{TYPEDEF}|{CONTROL}|{TYPE_QUAL}

ALLOC		new|delete
BOOL		bool|false|true
EXCEPTION	try|catch|throw
CLASS		class|public|protected|private|friend|explicit|mutable|inline|virtual|operator|this
CAST		static_cast|reinterpret_cast|const_cast|dynamic_cast
TEMPLATE	template|typename
NAMESPACE	namespace|using
TYPEID		typeid
CPP_KEYWORDS	{ALLOC}|{BOOL}|{EXCEPTION}|{CLASS}|{CAST}|{TEMPLATE}|{NAMESPACE}|{TYPEID}

DECLSPEC	__declspec|dllimport|dllexport|naked|thread
CALL_CONV	__cdecl|__stdcall|__fastcall
INT_TYPE	__int8|__int16|__int32|__int64
C_ENV_DEPEND	__asm|__inline|__forceinline|__try|__except|__finally|__based|__leave|{DECLSPEC}|{CALL_CONV}|{INT_TYPE}

MMREGISTER	__m64|__m128|__m128d|__m128i
CPP_DECLSPEC	deprecated|noinline|noreturn|nothrow|novtable|property|selectany|uuid
DOT_NET		__abstract|__box|__delegate|__gc|__pin|__property|__nogc|__sealed|__try_cast|__value
CPP_ENV_DEPEND	__alignof|__assume|__event|__hook|__identifier|__if_exist|__if_not_exist|__interface|__multiple_inheritance|__noop|__raise|__single_inheritance|__super|__unhook|__uuidof|__virtual_inheritance|__w64|{MMREGISTER}|{CPP_DECLSPEC}|{DOT_NET}

ENV_DEPEND	{C_ENV_DEPEND}|{CPP_ENV_DEPEND}

PRAGMA_KEYWORDS	warning|disable|enable|default|comment|lib|pack|push|pop

INTEGER		({NONZERO}{DIGIT}*|0)(u|I64)?
XINTEGER	0x{XDIGIT}+(u|I64)?
OINTEGER	0{DIGIT}+(u|I64)?
NUMBER		{FLOAT}{EXPONENT}?
KEYWORDS	{C_KEYWORDS}|{CPP_KEYWORDS}|{ENV_DEPEND}
ID		[_[:alpha:]][_[:alnum:]]*
WSPACE		[[:blank:]]+
RET		\r|\n|\r\n

%%

<<EOF>>		echof("</pre>"); return 0;

"//"		enter(LINE_COM); echof("<font color=darkgreen>//");
<LINE_COM>\\{RET}	ECHO;
<LINE_COM>{RET}	echof("</font>"); ECHO; leave();

"/*"		enter(COMMENT); echof("<font color=darkgreen>/*");
<COMMENT>"*/"	echof("*/</font>"); leave();
<COMMENT>"*"	ECHO;
<COMMENT>[^*]+	ECHO;

\"		enter(STRING); echof("<font color=darkcyan>\"");
<STRING>([^\"\r\n\\]|\\.|\\{RET})+	ECHO;
<STRING>{RET}	echof("</font>"); leave();
<STRING>\"	echof("\"</font>"); leave();

'		enter(CHARA); echof("<font color=darkcyan>'");
<CHARA>([^'\r\n]|\\.|\\{RET})+	ECHO;
<CHARA>{RET}	echof("</font>"); leave();
<CHARA>'	echof("'</font>"); leave();

<INCLUDE>"<"	enter(FILENAME); echof("<font color=darkcyan>&lt;");
<FILENAME>([^>\r\n]|\\[^>]|\\{RET})+	ECHO;
<FILENAME>\\">"	echof("\\&gt;");
<FILENAME>{RET}	echof("</font>"); ECHO; leave();
<FILENAME>">"	echof("&gt;</font>"); leave();

<PRAGMA>{PRAGMA_KEYWORDS}	echo_color("blue");
<PRAGMA>\\{RET}	ECHO;
<PRAGMA>{RET}	ECHO; leave();

{INTEGER}|{XINTEGER}|{OINTEGER}|{NUMBER}	echo_color("darkcyan");
{KEYWORDS}	echo_color("blue");
{WSPACE}	ECHO;
{ID}		ECHO;
;		ECHO;
"<"		echof("&lt;");
">"		echof("&gt;");
"&"		echof("&amp;");
{RET}		ECHO; if(YY_START == INCLUDE) leave();

#[[:blank:]]*include	enter(INCLUDE); echo_color("blue");
#[[:blank:]]*pragma	enter(PRAGMA); echo_color("blue");
#[[:blank:]]*{ID}	echo_color("blue");

.		ECHO;

%%

int main(int argc, char** argv)
{
  if(argc > 1 && strcmp(argv[1], "-") != 0)
    yyin  = fopen(argv[1], "r");
  if(argc > 2 && strcmp(argv[2], "-") != 0)
    yyout = fopen(argv[2], "w");

  echof("<pre>");
  stack_ptr = 0;
  return yylex();
}

static void echof(const char* format, ...)
{
  va_list args;
  va_start(args, format);
  vfprintf(yyout, format, args);
  va_end(args);
}

static void echo_color(const char* color)
{
  echof("<font color=%s>%s</font>", color, yytext);
}

static void enter(int cond)
{
  if(stack_ptr == CONDSTACK_SIZE)
  {
    fprintf(stderr, "Condition-stack overflow\n");
    exit(1);
  }
  cond_stack[stack_ptr++] = YY_START;
  BEGIN(cond);
}

static void leave(void)
{
  if(stack_ptr == 0)
    BEGIN(INITIAL);
  else
    BEGIN(cond_stack[--stack_ptr]);
}

